from .plot import plotScores
from .fit import fitModels
from .score import scoreModels, getBestModel, getBestScore
from .crossval import crossvalModels
from .parfit import bestFit

__all__ = (plot.__all__ + fit.__all__ + score.__all__ + crossval.__all__ + parfit.__all__)
__version__ = '0.210'



from joblib import Parallel, delayed
from sklearn.base import BaseEstimator
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

__all__ = ["fitModels"]


def fitOne(model, X, y, params):
    """
    Makes one model fit using provided data and parameters
    :param model: The instantiated model you wish to pass, e.g. LogisticRegression()
    :param X: The independent variable data
    :param y: The response variable data
    :param params: The parameters passed through to the model from the parameter grid
    :return: Returns the fitted model
    """
    if isinstance(model, BaseEstimator):
        model.set_params(**params)
    else:
        model = model(**params)
    return model.fit(X, y)


def fitModels(model, paramGrid, X, y, n_jobs=-1, verbose=10):
    """
    Parallelizes fitting all models using all combinations of parameters in paramGrid on provided data.
    :param model: The instantiated model you wish to pass, e.g. LogisticRegression()
    :param paramGrid: The ParameterGrid object created from sklearn.model_selection
    :param X: The independent variable data
    :param y: The response variable data
    :param n_jobs: Number of cores to use in parallelization (defaults to -1: all cores)
    :param verbose: The level of verbosity of reporting updates on parallel process
        Default is 10 (send an update at the completion of each job)
    :return: Returns a list of fitted models

    Example usage:
        from sklearn.linear_model import LogisticRegression
        from sklearn.model_selection import ParameterGrid
        model = LogisticRegression()
        grid = {
            'C': [1e-4, 1e-3], # regularization
            'penalty': ['l1','l2'], # penalty type
            'n_jobs': [-1] # parallelize within each fit over all cores
        }
        paramGrid = ParameterGrid(grid)
        myModels = fitModels(model, paramGrid, X_train, y_train)
    """
    return Parallel(n_jobs=n_jobs, verbose=verbose)(delayed(fitOne)(model, X, y, params) for params in paramGrid)



from sklearn.metrics import roc_auc_score
from .fit import fitModels
from .score import scoreModels, getBestScore, getBestModel
from .plot import plotScores
from .crossval import crossvalModels
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

__all__ = ["bestFit"]


def bestFit(
    model,
    paramGrid,
    X_train,
    y_train,
    X_val=None,
    y_val=None,
    nfolds=5,
    metric=roc_auc_score,
    greater_is_better=True,
    predict_proba=True,
    showPlot=True,
    scoreLabel=None,
    vrange=None,
    cmap="YlOrRd",
    n_jobs=-1,
    verbose=10,
):
    """
    Parallelizes choosing the best fitting model on the validation set, doing a grid search over the parameter space.
        Models are scored using specified metric, and user must determine whether the best score is the 'max' or 'min' of scores.
    :param model: The instantiated model you wish to pass, e.g. LogisticRegression()
    :param paramGrid: The ParameterGrid object created from sklearn.model_selection
    :param X_train: The independent variable data used to fit the models
    :param y_train: The dependent variable data used to fit the models
    :param X_val: The independent variable data used to score the models
    :param y_val: The dependent variable data used to score the models
    :param nfolds: Cross-validation number of folds, used if a validation set is not specified
    :param metric: The metric used to score the models, e.g. imported from sklearn.metrics
    :param greater_is_better: Choice between optimizing for greater scores or lesser scores
        Default True means greater and False means lesser
    :param predict_proba: Choice between 'predict_proba' and 'predict' for scoring routine
        Default True means predict_proba and False means predict
    :param showPlot: Whether or not to display the plot of the scores over the parameter grid
    :param scoreLabel: The specified label (dependent on scoring metric used), e.g. 'AUC'
    :param vrange: The visible range over which to display the scores
    :param cmap: The chosen colormap for 2D and 3D plotting. Default is YlOrRd.
        You can invert your chosen colormap by adding '_r' to the end
    :param n_jobs: Number of cores to use in parallelization (defaults to -1: all cores)
    :param verbose: The level of verbosity of reporting updates on parallel process
        Default is 10 (send an update at the completion of each job)
    :return: Returns a tuple including the best scoring model, the score of the best model, all models, and all scores
    """
    if (X_val is None) or (y_val is None):
        print("-------------CROSS-VALIDATING MODELS-------------")
        scores, models = crossvalModels(
            model, paramGrid, X_train, y_train, nfolds, metric, predict_proba, n_jobs, verbose
        )
    else:
        print("-------------FITTING MODELS-------------")
        models = fitModels(model, paramGrid, X_train, y_train, n_jobs, verbose)
        print("-------------SCORING MODELS-------------")
        scores = scoreModels(models, X_val, y_val, metric, predict_proba, n_jobs, verbose)
    if showPlot:
        plotScores(scores, paramGrid, scoreLabel, greater_is_better, vrange, cmap)

    return getBestModel(models, scores, greater_is_better), getBestScore(scores, greater_is_better), models, scores



import matplotlib.pyplot as plt
import matplotlib.colorbar as cb
import numpy as np
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

__all__ = ["plotScores"]


def plot1DGrid(scores, paramsToPlot, scoreLabel, vrange):
    """
    Makes a line plot of scores, over the parameter to plot
    :param scores: A list of scores, estimated using scoreModels
    :param paramsToPlot: The parameter to plot, chosen automatically by plotScores
    :param scoreLabel: The specified score label (dependent on scoring metric used)
    :param vrange: The yrange of the plot
    """
    key = list(paramsToPlot.keys())
    plt.figure(figsize=(int(round(len(paramsToPlot[key[0]]) / 1.33)), 6))
    plt.plot(np.linspace(0, len(paramsToPlot[key[0]]), len(scores)), scores, "-or")
    plt.xlabel(key[0])
    plt.xticks(np.linspace(0, len(paramsToPlot[key[0]]), len(scores)), paramsToPlot[key[0]])
    if scoreLabel is not None:
        plt.ylabel(scoreLabel)
    else:
        plt.ylabel("Score")
    if vrange is not None:
        plt.ylim(vrange[0], vrange[1])
    plt.box(on=False)
    plt.show()


def plot2DGrid(scores, paramsToPlot, keysToPlot, scoreLabel, greater_is_better, vrange, cmap):
    """
    Plots a heatmap of scores, over the paramsToPlot
    :param scores: A list of scores, estimated using parallelizeScore
    :param paramsToPlot: The parameters to plot, chosen automatically by plotScores
    :param scoreLabel: The specified score label (dependent on scoring metric used)
    :param greater_is_better: Choice between optimizing for greater scores or lesser scores
        Used to make better scores darker on colormap
        Default True means greater and False means lesser
    :param vrange: The visible range of the heatmap (range you wish the heatmap to be specified over)
    :param cmap: The chosen colormap for 2D and 3D plotting. Default is YlOrRd
    """
    scoreGrid = np.reshape(scores, (len(paramsToPlot[keysToPlot[0]]), len(paramsToPlot[keysToPlot[1]])))
    plt.figure(
        figsize=(
            int(round(len(paramsToPlot[keysToPlot[1]]) / 1.33)),
            int(round(len(paramsToPlot[keysToPlot[0]]) / 1.33)),
        )
    )
    if not greater_is_better:
        if cmap.endswith("_r"):
            cmap = cmap[:-2]
        else:
            cmap = cmap + "_r"
    if vrange is not None:
        plt.imshow(scoreGrid, cmap=cmap, vmin=vrange[0], vmax=vrange[1])
    else:
        plt.imshow(scoreGrid, cmap=cmap)
    plt.xlabel(keysToPlot[1])
    plt.xticks(np.arange(len(paramsToPlot[keysToPlot[1]])), paramsToPlot[keysToPlot[1]])
    plt.ylabel(keysToPlot[0])
    plt.yticks(np.arange(len(paramsToPlot[keysToPlot[0]])), paramsToPlot[keysToPlot[0]])
    if scoreLabel is not None:
        plt.title(scoreLabel)
    else:
        plt.title("Score")
    plt.colorbar()
    plt.box(on=False)
    plt.show()


def plot3DGrid(scores, paramsToPlot, keysToPlot, scoreLabel, greater_is_better, vrange, cmap):
    """
    Plots a grid of heatmaps of scores, over the paramsToPlot
    :param scores: A list of scores, estimated using parallelizeScore
    :param paramsToPlot: The parameters to plot, chosen automatically by plotScores
    :param scoreLabel: The specified score label (dependent on scoring metric used)
    :param greater_is_better: Choice between optimizing for greater scores or lesser scores
        Used to make better scores darker on colormap
        Default True means greater and False means lesser
    :param vrange: The visible range of the heatmap (range you wish the heatmap to be specified over)
    :param cmap: The chosen colormap for 2D and 3D plotting. Default is YlOrRd
    """
    vmin = np.min(scores)
    vmax = np.max(scores)
    scoreGrid = np.reshape(
        scores, (len(paramsToPlot[keysToPlot[0]]), len(paramsToPlot[keysToPlot[1]]), len(paramsToPlot[keysToPlot[2]]))
    )

    smallest_dim = np.argmin(scoreGrid.shape)
    if smallest_dim != 2:
        scoreGrid = np.swapaxes(scoreGrid, smallest_dim, 2)
        keysToPlot[smallest_dim], keysToPlot[2] = keysToPlot[2], keysToPlot[smallest_dim]

    nelements = scoreGrid.shape[2]
    nrows = np.floor(nelements ** 0.5).astype(int)
    ncols = np.ceil(1.0 * nelements / nrows).astype(int)
    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex="all",
        sharey="all",
        figsize=(
            int(round(len(paramsToPlot[keysToPlot[1]]) * ncols * 1.33)),
            int(round(len(paramsToPlot[keysToPlot[0]]) * nrows * 1.33)),
        ),
    )

    if not greater_is_better:
        if cmap.endswith("_r"):
            cmap = cmap[:-2]
        else:
            cmap = cmap + "_r"
    i = 0
    for ax in axes.flat:
        if vrange is not None:
            im = ax.imshow(scoreGrid[:, :, i], cmap=cmap, vmin=vrange[0], vmax=vrange[1])
        else:
            im = ax.imshow(scoreGrid[:, :, i], cmap=cmap, vmin=vmin, vmax=vmax)
        ax.set_xlabel(keysToPlot[1])
        ax.set_xticks(np.arange(len(paramsToPlot[keysToPlot[1]])))
        ax.set_xticklabels(paramsToPlot[keysToPlot[1]])
        ax.set_ylabel(keysToPlot[0])
        ax.set_yticks(np.arange(len(paramsToPlot[keysToPlot[0]])))
        ax.set_yticklabels(paramsToPlot[keysToPlot[0]])
        ax.set_title(keysToPlot[2] + " = " + str(paramsToPlot[keysToPlot[2]][i]))
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["left"].set_visible(False)
        i += 1
        if i == nelements:
            break
    if scoreLabel is not None:
        fig.suptitle(scoreLabel, fontsize=18)
    else:
        fig.suptitle("Score", fontsize=18)
    fig.subplots_adjust(right=0.8)
    cbar = cb.make_axes(ax, location="right", fraction=0.03)
    fig.colorbar(im, cax=cbar[0])
    plt.show()


def plotScores(scores, paramGrid, scoreLabel=None, greater_is_better=True, vrange=None, cmap="YlOrRd"):
    """
    Makes a plot representing how the scores vary over the parameter grid
        Automatically decides whether to use a simple line plot (varying over one parameter)
        or a heatmap (varying over two parameters)
    :param scores: A list of scores, estimated using scoreModels
    :param paramGrid: The parameter grid specified when fitting the models using fitModels
    :param scoreLabel: The specified label (dependent on scoring metric used), e.g. 'AUC'
    :param greater_is_better: Choice between optimizing for greater scores or lesser scores
        Used to make better scores darker on colormap
        Default True means greater and False means lesser
    :param vrange: The visible range over which to display the scores
    :param cmap: The chosen colormap for 2D and 3D plotting. Default is YlOrRd.
        You can invert your chosen colormap by adding '_r' to the end
    :return:
    """

    keys = sorted(list(paramGrid)[0].keys())
    uniqParams = dict()
    order = dict()
    for k in keys:
        order[k] = np.unique([str(params[k]) for params in list(paramGrid)], return_index=True)[1]
        uniqParams[k] = [params[k] for params in np.asarray(list(paramGrid))[sorted(order[k])]]

    keysToPlot = list()
    for k in keys:
        if len(uniqParams[k]) > 1:
            keysToPlot.append(k)

    for k in keys:
        if k not in keysToPlot:
            uniqParams.pop(k, None)

    numDim = len(keysToPlot)
    if numDim > 3:
        print("Too many dimensions to plot.")
    elif numDim == 3:
        plot3DGrid(scores, uniqParams, keysToPlot, scoreLabel, greater_is_better, vrange, cmap)
    elif numDim == 2:
        plot2DGrid(scores, uniqParams, keysToPlot, scoreLabel, greater_is_better, vrange, cmap)
    elif numDim == 1:
        plot1DGrid(scores, uniqParams, scoreLabel, vrange)
    else:
        print("No parameters that vary in the grid")



from joblib import Parallel, delayed
from sklearn.metrics import roc_auc_score
import numpy as np
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

__all__ = ["scoreModels", "getBestModel", "getBestScore"]


def scoreOne(model, X, y, metric, predict_proba):
    """
    Scores one model fit using provided metric for given model on scoring data.
    :param model: The fitted model you wish to score
    :param X: The dependent variable data you wish to use for prediction
    :param y: The ground truth independent variable data you wish to compare the predictions to
    :param metric: The metric you wish to use to score the predictions using
    :param predict_proba: Choice between using 'predict_proba' and 'predict' for scoring routine.
        Default True means predict_proba and False means predict
    :return: Returns the score
    """
    if predict_proba:
        try:
            return metric(y, model.predict_proba(X)[:, 1])
        except:
            print("This model/metric cannot use predict_proba. Using predict for scoring instead.")
            return metric(y, model.predict(X))
    else:
        return metric(y, model.predict(X))


def scoreModels(models, X, y, metric=roc_auc_score, predict_proba=True, n_jobs=-1, verbose=10):
    """
    Parallelizes scoring all models using provided metric for given models on scoring data
    :param models: The lists of fitted models you wish to score, fitted using fitModels
    :param X: The X data ou wish to use for prediction
    :param y: The ground truth y data you wish to compare the predictions to
    :param metric: The metric you wish to use to score the predictions using
        Defaults to roc_auc_score
    :param predict_proba: Choice between 'predict_proba' and 'predict' for scoring routine
        Default True means predict_proba and False means predict
    :param n_jobs: Number of cores to use in parallelization (defaults to -1: all cores)
    :param verbose: The level of verbosity of reporting updates on parallel process
        Default is 10 (send an update at the completion of each job)
    :return: Returns a list of scores in the same order as the list of models

    Example usage:
        from sklearn.metrics import recall_score
        myScores = scoreModels(myModels, X_val, y_val, recall_score)

    """
    return Parallel(n_jobs=n_jobs, verbose=np.ceil(verbose / 10))(
        delayed(scoreOne)(m, X, y, metric, predict_proba) for m in models
    )


def getBestModel(models, scores, greater_is_better=True):
    """
    Returns the best model from the models list based on the scores from
    the scores list. "Best" means 'max' or 'min' of scores, dependent on greater_is_better
    :param models: List of models returned by fitModels or parameters returned by crossvalModels
    :param scores: List of corresponding scores returned by scoreModels
    :param greater_is_better: Choice between optimizing for greater scores or lesser scores
        Default True means greater and False means lesser
    :return: The best model, with associated hyper-parameters
    """
    if greater_is_better:
        return models[np.argmax(scores)]
    else:
        return models[np.argmin(scores)]


def getBestScore(scores, greater_is_better=True):
    """
    Returns the score of the best model from the models list based on the scores from
    the scores lsit. "Best" means 'max' or 'min' of scores, dependent on greater_is_better
    :param scores: List of corresponding scores returned by scoreModels
    :param greater_is_better: Choice between optimizing for greater scores or lesser scores
        Default True means greater and False means lesser
    :return: The score of the best model
    """
    if greater_is_better:
        return np.max(scores)
    else:
        return np.min(scores)



from distutils.core import setup
setup(
    name='parfit',
    packages=['parfit'],  # this must be the same as the name above
    version='0.210',
    description='A package for parallelizing the fit and flexibly scoring of sklearn machine learning models, with optional plotting routines.',
    author='Jason Carpenter',
    author_email='jmcarpenter2@dons.usfca.edu',
    url='https://github.com/jmcarpenter2/parfit',  # use the URL to the github repo
    download_url='https://github.com/jmcarpenter2/parfit/archive/0.210.tar.gz',
    keywords=['parallelize', 'sklearn', 'machinelearning',
              'models', 'fit'],  # keywords
    install_requires=[
        'joblib',
        'matplotlib',
        'sklearn',
        'numpy'
    ],
    classifiers=[],
)



from joblib import Parallel, delayed
from sklearn.base import BaseEstimator
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import roc_auc_score
import numpy as np
from .fit import fitOne
from .score import scoreOne
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


__all__ = ["crossvalModels"]


def crossvalOne(model, X, y, params, nfolds, metric=roc_auc_score, predict_proba=True, n_jobs=-1, verbose=1):
    """
    Makes one cross-validation model fit-score run using provided data and parameters
    :param model: The instantiated model you wish to pass, e.g. LogisticRegression()
    :param X: The independent variable data
    :param y: The response variable data
    :param params: The parameters passed through to the model from the parameter grid
    :param nfolds: The number of folds you wish to use for cross-validation
    :param metric: The metric you wish to use to score the crossval predictions using
    :param predict_proba: Choice between using 'predict_proba' and 'predict' for scoring routine.
        Default True means predict_proba and False means predict
    :param n_jobs: Number of cores to use in parallelization (defaults to -1: all cores)
    :param verbose: The level of verbosity of reporting updates on parallel process
        Default is 10 (send an update at the completion of each job)
    :return: Returns the mean of the cross-validation scores
    """
    if 'random_state' in params.keys():
        random_state = params['random_state']
    else:
        random_state = None
    kf = StratifiedKFold(n_splits=nfolds, random_state=random_state)
    train_indices, test_indices = zip(*kf.split(X))
    if isinstance(model, BaseEstimator):
        fitted_models = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(fitOne)(type(model), np.asarray(X)[train_index], np.asarray(y)[train_index], params)
            for train_index in train_indices
        )
    else:
        fitted_models = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(fitOne)(model, np.asarray(X)[train_index], np.asarray(y)[train_index], params)
            for train_index in train_indices
        )

    scores = Parallel(n_jobs=n_jobs, verbose=0)(
        delayed(scoreOne)(fitted_model, np.asarray(X)[test_index], np.asarray(y)[test_index], metric, predict_proba)
        for fitted_model, test_index in zip(fitted_models, test_indices)
    )

    return np.mean(scores)


def crossvalModels(model, paramGrid, X, y, nfolds=5, metric=roc_auc_score, predict_proba=True, n_jobs=-1, verbose=10):
    """
    Parallelizes fitting and scoring all cross-validation models using all combinations of parameters in paramGrid on provided data.
    :param model: The instantiated model you wish to pass, e.g. LogisticRegression()
    :param paramGrid: The ParameterGrid object created from sklearn.model_selection
    :param X: The independent variable data
    :param y: The response variable data
    :param nfolds: The number of folds you wish to use for cross-validation
    :param metric: The metric you wish to use to score the crossval predictions using
    :param predict_proba: Choice between using 'predict_proba' and 'predict' for scoring routine.
        Default True means predict_proba and False means predict
    :param n_jobs: Number of cores to use in parallelization (defaults to -1: all cores)
    :param verbose: The level of verbosity of reporting updates on parallel process
        Default is 10 (send an update at the completion of each job)
    :return: Returns the grid of mean of cross-validation scores for the specified parameters,
        and the associated paramGrid
    """
    return (
        Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(crossvalOne)(model, X, y, params, nfolds, metric, predict_proba, n_jobs, np.ceil(verbose / 10))
            for params in paramGrid
        ),
        list(paramGrid),
    )
